#!/usr/bin/env python3
# -*- coding: utf-8 -*-
r"""
Life's pathetic, have fun ("▔□▔)/hi~♡ Nasy.

Excited without bugs::

    |             *         *
    |                  .                .
    |           .
    |     *                      ,
    |                   .
    |
    |                               *
    |          |\___/|
    |          )    -(             .              ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Feb 15, 2019
email    : Nasy <nasyxx+nalude@gmail.com>
filename : extra.py
project  : nalude
license  : LGPL-3.0+

There are more things in heaven and earth, Horatio, than are dreamt.
 --  From "Hamlet"
"""
# Standard Library
from typing import Type, Tuple, Iterable

# Local Packages
from .types import G, a


def flatten(xs: a, *, ignore: Tuple[Type, ...] = (dict,)) -> G:
    """Flatten iterables of iterable to a single one.

    The true type of it is:: `T = Iterable[Untion[a, T]]` ."""
    if not (isinstance(xs, (str, *ignore)) or not isinstance(xs, Iterable)):
        for x in xs:
            for x_ in flatten(x, ignore=ignore):
                yield x_
    else:
        yield xs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
r"""
Life's pathetic, have fun ("▔□▔)/hi~♡ Nasy.

Excited without bugs::

    |             *         *
    |                  .                .
    |           .
    |     *                      ,
    |                   .
    |
    |                               *
    |          |\___/|
    |          )    -(             .              ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Jan 31, 2019
email    : Nasy <nasyxx+nalude@gmail.com>
filename : nalude.py
project  : nalude
license  : LGPL-3.0+

There are more things in heaven and earth, Horatio, than are dreamt.
 --  From "Hamlet"
"""
# Standard Library
from typing import List, Tuple, Callable, Iterable, Generator
from functools import reduce

# Local Packages
from .types import G, Num, Predicate, a, b, c

# ----------------------------------------------------------------------
# Folds and Traversals
# ----------------------------------------------------------------------


def foldl(f: Callable[[b, a], b], init: b, t: Iterable[a]) -> b:
    """Left-associative fold of an iterable."""
    return reduce(f, t, init)


def foldr(f: Callable[[a, b], b], init: b, t: Iterable[a]) -> b:
    """Right-associative fold of an iterable."""
    t_ = list(t)
    return reduce(flip(f), reversed(t_), init)


def foldl1(f: Callable[[a, a], a], t: Iterable[a]) -> a:
    """Left-associative fold of an iterable.

    This is avariant of foldl without base case.
    """
    t_ = iter(t)
    try:
        return reduce(f, t_, next(t_))
    except StopIteration:
        raise IndexError("Iterable is Empty")


def foldr1(f: Callable[[a, a], a], t: Iterable[a]) -> a:
    """Right-associative fold of an iterable.

    This is a variant of foldr without base case.
    """
    t_ = iter(reversed(list(t)))
    try:
        return reduce(f, t_, next(t_))
    except StopIteration:
        raise IndexError("Iterable is Empty")


def product(nums: Iterable[Num]) -> Num:
    """Computes the product of the numbers."""
    try:
        return foldl1(lambda b, a: a * b, nums)
    except IndexError:
        raise IndexError("Numbers are empty")


# ----------------------------------------------------------------------
# Lists
# ----------------------------------------------------------------------


def head(xs: Iterable[a]) -> a:
    """Extract the first element of an iterable."""
    try:
        return next(iter(xs))
    except StopIteration:
        raise IndexError("Iterable is empty.")


def tail(xs: Iterable[a]) -> G:
    """Extract the elements after the head of an iterable."""
    try:
        ixs = iter(xs)
        next(ixs)
        for x in ixs:
            yield x
    except StopIteration:
        raise IndexError("Iterable is empty.")


def last(xs: Iterable[a]) -> a:
    """Extract the last element of an iterable."""
    try:
        return next(iter(reversed(list(xs))))
    except StopIteration:
        raise IndexError("Iterable is empty.")


def init(xs: Iterable[a]) -> G:
    """Extract all the elements of an iterable except the last one."""
    ixs = iter(xs)
    try:
        x_ = next(ixs)
        n = True
        while n:
            try:
                x = next(ixs)
                yield x_
                x_ = x
            except StopIteration:
                n = False
    except StopIteration:
        raise IndexError("Iterable is empty.")


def null(l: Iterable[a]) -> bool:
    """Test whether the sequence is empty."""
    try:
        next(iter(l))
        return False
    except StopIteration:
        return True


def iterate(f: Callable[[a], a], x: a) -> G:
    """Yield a tuple of repeated applications of f to x."""
    while True:
        yield x
        x = f(x)


def repeat(x: a) -> G:
    """Repeat x."""
    while True:
        yield x


def replicate(n: int, x: a) -> G:
    """Return a iterator of length n with x the value of every element."""
    for _ in range(n):
        yield x


def cycle(xs: Iterable[a]) -> G:
    """Tie an iterable to infinite circuler one."""
    xs_ = []  # if x is a finite iterator, this can save the value.
    ixs = iter(xs)
    try:
        x = next(ixs)
        yield x
        xs_.append(x)
    except StopIteration:
        raise IndexError("Iterable is empty.")
    for x in ixs:
        yield x
        xs_.append(x)
    while True:
        for x in xs_:
            yield x


def take(n: int, xs: Iterable[a]) -> G:
    """Return the first n elements of sequence xs."""
    for _, x in zip(range(n), xs):
        yield x


def drop(n: int, xs: Iterable[a]) -> G:
    """Return the remaining elements of xs after the first n elements."""
    ixs = iter(xs)
    try:
        for _ in range(n):
            next(ixs)
        while True:
            yield next(ixs)
    except StopIteration:
        pass


def splitat(n: int, xs: Iterable[a]) -> Tuple[G, G]:
    """Equal to (take(n, xs), drop(n, xs)).

    Return a tuple where the first element is xs prefix of length n and second
    element is the remainder of the sequence xs.
    """
    ixs = iter(xs)
    return (i for i in list(take(n, ixs))), (i for i in ixs)


def takewhile(p: Predicate, xs: Iterable[a]) -> G:
    """Return the longest prefix of xs of elements that satisfy predicate p."""
    for x in xs:
        if p(x):
            yield x
        else:
            break


def dropwhile(p: Predicate, xs: Iterable[a]) -> G:
    """Returns the suffix remaining after takewhile(p, xs)."""
    ixs = iter(xs)
    for x in ixs:
        if not p(x):
            yield x
            break
    for x in ixs:
        yield x


def span(p: Predicate, xs: Iterable[a]) -> Tuple[List[a], List[a]]:
    """Equal to (takewhile(p, xs), dropwhile(p, xs)).

    Note: Not lazy."""
    lxs = list(xs)
    return list(takewhile(p, lxs)), list(dropwhile(p, lxs))


def break_(p: Predicate, xs: Iterable[a]) -> Tuple[List[a], List[a]]:
    """Equal to (takewhile(not_(p), xs), dropwhile(not_(p), xs)).

    Note: Not lazy."""
    lxs = list(xs)
    return list(takewhile(not_(p), lxs)), list(dropwhile(not_(p), lxs))


# ----------------------------------------------------------------------
# Miscellaneous
# ----------------------------------------------------------------------


def id_(x: a) -> a:
    """Identity function."""
    return x


def const(x: a, _: b) -> a:
    """Evaluates to x for all inputs."""
    return x


def o(f1: Callable[[b], c], f2: Callable[[a], b]) -> Callable[[a], c]:
    """Function composition.

    >>> o(f1, f2)(a) == f1(f2(a))
    """
    return lambda a_: f1(f2(a_))


def flip(f: Callable[[a, b], c]) -> Callable[[b, a], c]:
    """Flip takes its two arguments into the reverse order of f.

    >>> flip(f)(b, a) == f(a, b)
    """
    return lambda b_, a_: f(a_, b_)


def until(p: Predicate, f: Callable[[a], a], x: a) -> a:
    """Yields the result of applying f until p holds."""
    return p(x) and x or until(p, f, f(x))


# ----------------------------------------------------------------------
# Strings
# ----------------------------------------------------------------------


def lines(s: str) -> List[str]:
    """Break up a string into a list of strings at newline characters."""
    return s.splitlines()


def unlines(xs: Iterable[str]) -> str:
    """The inverse operation of lines, append a newline to each."""
    return "\n".join(xs) + ("\n" if xs else "")


def words(s: str) -> List[str]:
    """Breaks a string up into a list of words, which were delimited by white
    space."""
    return s.split()


def unwords(xs: Iterable[str]) -> str:
    """The inverse operation of words, join words with space."""
    return " ".join(xs)


# ----------------------------------------------------------------------
# Specials
# ----------------------------------------------------------------------


def not_(f: Callable[..., bool]) -> Callable[..., bool]:
    """Boolean "not"."""

    def _wrap(*args: a, **kwds: b) -> bool:
        """Wrapper of not_."""
        return not f(*args, **kwds)

    return _wrap


def all_(p: Predicate, xs: Iterable[a]) -> bool:
    """Determine whether all elements of the structure satisfy the p."""
    return all(map(p, xs))


def any_(p: Predicate, xs: Iterable[a]) -> bool:
    """Determine whether any element of the structure satisfies the p."""
    return any(map(p, xs))


def concat(xss: Iterable[Iterable[a]]) -> G:
    """The concatenation of all the elements of a container of Iterables."""
    for xs in xss:
        for x in xs:
            yield x


def concatmap(
    f: Callable[[Iterable[a]], Iterable[b]], xss: Iterable[Iterable[a]]
) -> Generator[b, None, None]:
    """Map a function over all the elements of a container and concatenate the
    resulting lists."""
    for xs in xss:
        for x in f(xs):
            yield x


# ----------------------------------------------------------------------
# Tuples
# ----------------------------------------------------------------------


def fst(t: Tuple[a, b]) -> a:
    """Extract the first component of a tuple."""
    return t[0]


def snd(t: Tuple[a, b]) -> b:
    """Extract the second component of a tuple."""
    return t[1]


def curry(f: Callable[[Tuple[a, b]], c], a_: a, b_: b) -> c:
    """Converts an uncurried function to a curried function."""
    return f((a_, b_))


def uncurry(f: Callable[[a, b], c], ab: Tuple[a, b]) -> c:
    """Converts a curried function to a function on pairs."""
    a_, b_ = ab
    return f(a_, b_)


# ----------------------------------------------------------------------
# Zip and Unzip
# ----------------------------------------------------------------------


def zipwith(
    f: Callable[..., c], *seqs: Iterable[a]
) -> Generator[c, None, None]:
    """Zipwith is map(f, zip), but f accept separate args instead of tuple"""
    for zseq in zip(*seqs):
        yield f(*zseq)


def unzip(pairs: Iterable[Tuple[a, ...]]) -> Tuple[Iterable[a], ...]:
    """Transform an iterable of pairs into a tuple of sequence.

    Note: Not lazy."""
    pairs = list(pairs)
    if pairs:
        return tuple(
            [pair[i] for pair in pairs] for i in range(len(head(pairs)))
        )
    return tuple()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
r"""
Life's pathetic, have fun ("▔□▔)/hi~♡ Nasy.

Excited without bugs::

    |             *         *
    |                  .                .
    |           .
    |     *                      ,
    |                   .
    |
    |                               *
    |          |\___/|
    |          )    -(             .              ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Jan 30, 2019
email    : Nasy <nasyxx+nalude@gmail.com>
filename : __init__.py
project  : nalude
license  : LGPL-3.0+

There are more things in heaven and earth, Horatio, than are dreamt.
 --  From "Hamlet"
"""
__version__ = "0.3.0"
# Local Packages
from .extra import flatten
from .nalude import (o, fst, id_, snd, all_, any_, drop, flip, head, init,
                     last, not_, null, span, tail, take, const, curry,
                     cycle, foldl, foldr, lines, until, unzip, words,
                     break_, concat, foldl1, foldr1, repeat, iterate,
                     product, splitat, uncurry, unlines, unwords, zipwith,
                     concatmap, dropwhile, replicate, takewhile,)

if __name__ == "__main__":
    # Folds and Traversals
    assert foldl
    assert foldl1
    assert foldr
    assert foldr1
    assert product

    # Lists
    assert head
    assert last
    assert init
    assert tail
    assert null
    assert iterate
    assert repeat
    assert replicate
    assert cycle
    assert take
    assert drop
    assert splitat
    assert takewhile
    assert dropwhile
    assert span
    assert break_

    # Miscellaneous
    assert id_
    assert const
    assert flip
    assert o
    assert until

    # String
    assert lines
    assert unlines
    assert words
    assert unwords

    # Specials
    assert not_
    assert all_
    assert any_
    assert concat
    assert concatmap

    # TUple
    assert fst
    assert snd
    assert curry
    assert uncurry

    # Zip and Unzip
    assert zipwith
    assert unzip

    # Extra
    assert flatten



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['nalude']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'nalude',
    'version': '0.3.0',
    'description': "A standard module.  Inspired by Haskell's Prelude.",
    'long_description': '# Table of Contents\n\n1.  [Prologue](#org8b6bf34)\n2.  [Nalude](#org0812938)\n3.  [Structure of Functions](#org658c0e2)\n    1.  [Folds and Traversals](#orgebe257d)\n    2.  [Lists](#org8fd809b)\n    3.  [Miscellaneous](#orgd3b3bf1)\n    4.  [String](#org8147070)\n    5.  [Specials](#org2de95e8)\n    6.  [Tuples](#orge6105ca)\n    7.  [Zip and Unzip](#orgcdca5f7)\n    8.  [Extra](#org5b1da39)\n4.  [Get Start](#orgccf98f8)\n    1.  [Install](#org1d1b386)\n5.  [Epoligue](#orgd300cef)\n    1.  [History](#org1d83dae)\n        1.  [Version 0.3.0](#orga174a69)\n        2.  [Version 0.2.0](#orga9ce38e)\n        3.  [Version 0.1.0](#orgad971e1)\n\n\n\n<a id="org8b6bf34"></a>\n\n# Prologue\n\nI like Haskell, also hope some of the prelude functions in Haskell can be used in Python.\n\nWish you enjoy coding with nalude.\n\n\n<a id="org0812938"></a>\n\n# Nalude\n\nNalude is a standard module, which is inspired by Haskell\'s Prelude.  Thus, nalude is also a\nlibrary of functional programming.\n\nNalude doesn\'t explicitly distinguish between Applicative\'s or monadic function since they are not\nwell differentiated in the basic structure of Python.  Of course, the main reason is that I am too\nfresh.\n\nThe following functions, Nalude tries to implement:\n\n1.  Folds and Traversals\n2.  Lists\n3.  Miscellaneous\n4.  Specials\n5.  Strings\n6.  Tuples\n7.  Zip and Unzip\n\nAt the same time, it also implements some functions which are only available in Python and are\nincluded in \\`extra\\`.\n\n\n<a id="org658c0e2"></a>\n\n# Structure of Functions\n\n\n<a id="orgebe257d"></a>\n\n## Folds and Traversals\n\n-   **foldr(f: Callable[[b, a], b], init: b, t: Sequence[a]):** Right-associative fold of a sequence.\n-   **foldl(f, init, t):** Left-associative fold of an iterable.  The same as `reduce`.\n-   **foldr1(f, t):** A variant of foldr without base case.\n-   **foldl1(f, t):** A variant of foldl without base case.\n-   **product(nums):** Computes the product of the numbers.\n-   **(HELP WANTED) traverse:** Map each element of a structure to an action, evaluate these actions\n    from left to right, and collect the results.\n-   **(HELP WANTED) sequence:** Evaluate each actions in the structure from left to right, and collect\n    the results.\n\n\n<a id="org8fd809b"></a>\n\n## Lists\n\n-   **head(xs):** Extract the first element of an iterable.\n-   **last(xs):** Extract the last element of an iterable.\n-   **null(xs):** Test whether the sequence is empty.\n-   Infinite lists\n    -   **iterate(f, x):** Yield a tuple of repeated applications of f to x.\n    -   **repeat(x):** Repeat x.\n    -   **replicate(n, x):** Return a list of length n with x the value of every element.\n    -   **cycle(xs):** Tie an iterable to infinite circuler one.\n\n    -   Sublists\n        -   **tail(xs):** Extract the elements after the head of a list.\n        -   **init(xs):** Extract all the elements of an iterable except the last one.\n        -   **take(n, xs):** Return the first n elements of sequence xs.\n        -   **drop(n, xs):** Return the remaining elements of xs after the first n elements.\n        -   **splitat(n, xs):** Return a tuple where the first element is xs prefix of length n and\n            second element is the remainder of the sequence xs.\n        -   **takewhile(p, xs):** Return the longest prefix of xs of elements that satisfy predicate p.\n        -   **dropwhile(p, xs):** Return the suffix remaining after takewhile(p, xs).\n        -   **(LAZY ONE HELP WANTED) span(p, xs):** Equal to (takewhile(p, xs), dropwhile(p, xs)).\n        -   **(LAZY ONE HELP WANTED) break\\_(p, xs):** Equal to (takewhile(not\\_(p), xs),\n            dropwhile(not\\_(p), xs)).\n\n\n<a id="orgd3b3bf1"></a>\n\n## Miscellaneous\n\n-   **id\\_(x):** Identity function.\n-   **const(x, \\_):** Evaluates to x for all inputs.\n-   **o(f1, f2):** (.) Function composition\n-   **flip(f):** Flip takes its two arguments into the reverse order of f.\n-   **until(p, f, x):** Yield the result of applying f until p holds.\n\n\n<a id="org8147070"></a>\n\n## String\n\n-   **lines(s):** Break up a string into a list of strings at newline characters.\n-   **unlines(xs):** The inverse operation of lines, append a newline to each.\n-   **words(s):** Break a string up into a list of words, which were delimited by white space.\n-   **unwords(xs):** The inverse operation of words, join words with space.\n\n\n<a id="org2de95e8"></a>\n\n## Specials\n\n-   **not\\_(f):** Boolean "not".\n-   **all\\_(p, xs):** Determine whether all elements of the structure satisfy the p.\n-   **any\\_(p, xs):** Determine whether sny elements of the structure satisfies the p.\n-   **concat(xss):** The concatenation of all the elements of a container of Sequences.\n-   **concatmap(f, xss):** Map a function over all the elements of a container and concatenate the\n    resulting lists.\n\n\n<a id="orge6105ca"></a>\n\n## Tuples\n\n-   **fst(t):** Extract the first component of a tuple.\n-   **snd(t):** Extract the second component of a tuple.\n-   **curry(f, a, b):** Converts an uncurried function to a curried function.\n-   **uncurry(f, ab@(a, b)):** Converts a curried function to a function on pairs.\n\n\n<a id="orgcdca5f7"></a>\n\n## Zip and Unzip\n\n-   **zipwith(f, \\*seqs):** Zipwith is map(f, zip), but f accept separate args instead of tuple\n-   **unzip(pairs):** Transform an iterable of pairs into a tuple of sequence. (Not lazy)\n\n\n<a id="org5b1da39"></a>\n\n## Extra\n\n-   **flatten(xs,\\*,ignore=(dict,)):** Flatten iterables of iterable to a single iterable.  It will\n    ignore instances in `ignore` tuple.\n\n\n<a id="orgccf98f8"></a>\n\n# Get Start\n\n\n<a id="org1d1b386"></a>\n\n## Install\n\n    pip install nalude\n\n\n<a id="orgd300cef"></a>\n\n# Epoligue\n\n\n<a id="org1d83dae"></a>\n\n## History\n\n\n<a id="orga174a69"></a>\n\n### Version 0.3.0\n\n-   **Data:** <span class="timestamp-wrapper"><span class="timestamp">&lt;Sun Mar 10, 2019&gt;</span></span>\n-   **Add:** -   Custom ignore when flatten iterable.\n\n\n<a id="orga9ce38e"></a>\n\n### Version 0.2.0\n\n-   **Data:** <span class="timestamp-wrapper"><span class="timestamp">&lt;Fri Feb 15, 2019&gt;</span></span>\n-   **Add:** -   flatten in extra.\n\n\n<a id="orgad971e1"></a>\n\n### Version 0.1.0\n\n-   **Data:** <span class="timestamp-wrapper"><span class="timestamp">&lt;Wed Feb 06, 2019&gt;</span></span>\n-   **Commemorate Version:** First Release.\n',
    'author': 'Nasy',
    'author_email': 'nasyxx+nalude@gmail.com',
    'url': 'https://github.com/nasyxx/nalude',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
